      MODULE mpi_f08_compile_constants
      INTEGER :: mpi_byte = 0
      INTEGER :: mpi_comm_world = 0
      INTEGER :: mpi_integer = 0
      INTEGER :: mpi_integer4 = 0
      INTEGER :: mpi_integer8 = 0
      END MODULE
