/*
 * (C) Copyright 2024- ECMWF.
 *
 * This software is licensed under the terms of the Apache Licence Version 2.0
 * which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
 * In applying this licence, ECMWF does not waive the privileges and immunities
 * granted to it by virtue of its status as an intergovernmental organisation
 * nor does it submit to any jurisdiction.
 */

#ifndef FIAT_DRHOOK_ROCTX_DR_HOOK_ROCTX_H
#define FIAT_DRHOOK_ROCTX_DR_HOOK_ROCTX_H

void dr_hook_roctx_start (const char* name);
void dr_hook_roctx_end ();

#endif //FIAT_DRHOOK_ROCTX_DR_HOOK_ROCTX_H
